/***********************************************************************************
 * QGLE - A Graphical Interface to GLE                                             *
 * Copyright (C) 2006  A. S. Budden & J. Struyf                                    *
 *                                                                                 *
 * This program is free software; you can redistribute it and/or                   *
 * modify it under the terms of the GNU General Public License                     *
 * as published by the Free Software Foundation; either version 2                  *
 * of the License, or (at your option) any later version.                          *
 *                                                                                 *
 * This program is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   *
 * GNU General Public License for more details.                                    *
 *                                                                                 *
 * You should have received a copy of the GNU General Public License               *
 * along with this program; if not, write to the Free Software                     *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. *
 *                                                                                 *
 * Also add information on how to contact you by electronic and paper mail.        *
 ***********************************************************************************/

#include <QtCore>
#include <QtNetwork>
#include <QtDebug>

#include "serverthread.h"
#include "drawingobject.h"
#include "math.h"
#include "qgs.h"
#define DIR_SEP "/"
#include "mainwindow.h"
#include "qgle_statics.h"

GLEServerThread::GLEServerThread(QObject *parent, quint16 port)
	: QThread(parent)
{
	// Just save the requested port
	this->port = port;

	listening = false;
	updateDPI = false;
	isRendering = false;
	dpi = 0;
}

GLEServerThread::~GLEServerThread()
{
	// Use a locker to get access to shared variables,
	// set abort=true and wait for the server to die
	mutex.lock();
	abort = true;
	mutex.unlock();

	wait();

}

void GLEServerThread::setDPI(double new_dpi)
{
	if (dpi != new_dpi)
	{
		dpi = new_dpi;
		emit dpiChanged(dpi);
	}
}

void GLEServerThread::setAutoScale(bool ascale)
{
	updateDPI = ascale;
}

void GLEServerThread::setUserAutoScale(bool ascale, const QSize& areaSize)
{
	drawingAreaSize = areaSize;
	userAutoScale = ascale;
}

bool GLEServerThread::shouldAutoScale()
{
	return updateDPI;
}

void GLEServerThread::setDefaultResolution(double new_dpi)
{
	defaultDPI = new_dpi;
}

double GLEServerThread::defaultResolution()
{
	return(defaultDPI);
}

void GLEServerThread::initialise()
{
	// Make sure abort is safe
	abort = false;

	// Start the processing
	start(LowPriority);
}

bool GLEServerThread::isListening()
{
	return(listening);
}

void GLEServerThread::waitForServer()
{
	serverReadyMutex.lock();
	if (isListening()) {
		serverReadyMutex.unlock();
		return;
	}
	serverReady.wait(&serverReadyMutex);
	serverReadyMutex.unlock();
}

QString& GLEServerThread::getGhostScriptOutput()
{
	return GhostScriptOutput;
}

QString& GLEServerThread::getGhostScriptError()
{
	return GhostScriptError;
}

void GLEServerThread::run()
{
	// This is the bit that runs in a different thread

	// Initialise variables
	QTcpSocket *clientConnection;
	QString gleCommand;
	char buffer[2000];
	qint64 bytesread;
	QStringList parts;
	QString dirname, filename, fullname, part;
	QString glefname;

	QStringList gsargs;

	QImage gs_image;

	// Set up some handy regexps
	QRegExp rxdir(".*dir: *\"([^\"]*)\".*");
	QRegExp rxfile(".*epsfile: *\"([^\"]*)\".*");
	QRegExp glefile(".*glefile: *\"([^\"]*)\".*");
	QRegExp rxdpi(".*dpi: *\"([^\"]*)\".*");
	// This one checks whether the call was made by qgle
	QRegExp rxqgle(".*qgle: *\"true\".*");

	// Is the server connection from QGLE or from GLE?
	bool calledByQgle;

	// Create a new server and listen on the requested port
	server = new QTcpServer();
	if (!server->listen(QHostAddress::LocalHost, port))
	{
		emit serverMessage(tr("Unable to start the server"));
		serverReady.wakeAll();
		return;
	}

	serverReadyMutex.lock();
	listening = true;
	serverReady.wakeAll();
	serverReadyMutex.unlock();

	// Notify the user that it's running
	emit serverMessage(tr("Server Listening"));

	// Loop forever
	forever
	{
		// Check whether we should quit
		mutex.lock();
		if (abort) {
			mutex.unlock();
			server->close();
			delete server;
			listening = false;
			return;
		}
		mutex.unlock();


		// Wait for a new connection for 10ms
		if (server->waitForNewConnection(10))
		{
			// Get the connection
			clientConnection = server->nextPendingConnection();

			gleCommand = "";
			calledByQgle = false;

			// While we are connected
			while(clientConnection->state() == QAbstractSocket::ConnectedState)
			{
				// Check whether we should quit
				mutex.lock();
				if (abort)
				{
					listening = false;
					delete server;
					return;
				}
				mutex.unlock();

				// See whether there's anything to read
				if(clientConnection->waitForReadyRead(10))
				{
					// There is, so find out how much
					bytesread = clientConnection->bytesAvailable();
					// and read it
					clientConnection->read(buffer, bytesread);
					buffer[bytesread] = 0;
					// append it to our gleCommand buffer
					gleCommand += buffer;
				}
			}

			// Break it up into separate lines
			parts = gleCommand.split(QRegExp("[\\r\\n]+"));

			// Find the bits we're interested in
			bool hasDPISpecification = false;
			foreach(part, parts)
			{
				qDebug() << "Received: " << part;
				if(rxdir.exactMatch(part))
				{
					dirname = rxdir.capturedTexts()[1];
				}
				if(rxfile.exactMatch(part))
				{
					filename = rxfile.capturedTexts()[1];
				}
				if(rxdpi.exactMatch(part))
				{
					hasDPISpecification = true;
					setDPI(rxdpi.capturedTexts()[1].toDouble());
				}
				if(glefile.exactMatch(part))
				{
					glefname = glefile.capturedTexts()[1];
				}
				if(rxqgle.exactMatch(part))
				{
					calledByQgle = true;
				}
			}

			// derive the full name and notify the user
			fullname = dirname + QDir::separator() + filename;
			mutex.lock();

			bool isReload = true;

			// if new file, then existing drawn objects should be removed
			// note: this happens if you do "gle -p f1.gle" and then "gle -p f2.gle"
			QString newGleFile = dirname + QDir::separator() + glefname;
			if (((GLEMainWindow*)parent())->getCurrentGleFile() != newGleFile) {
				isReload = false;
			}

			gleFile.setEpsFile(fullname);
			gleFile.setGleFile(newGleFile);

			mutex.unlock();

			emit gleMinusPRunned(newGleFile, isReload);
			emit serverMessage(((QString) "Open: ") + newGleFile);

			// Delete the connection and loop round to wait for a new one
			delete(clientConnection);
		}
	}
}

void GLEServerThread::render(int caller, GLEDrawingObject *obj)
{
	int b1,b2,b3,b4;
	double width;
	double height;

	// Bounding box can have negative numbers
	QRegExp rxbb("^%%BoundingBox: +([\\-\\d]+) +([\\-\\d]+) +([\\-\\d]+) +([\\-\\d]+).*$");

	b1 = b2 = b3 = b4 = 0;

	GSInterpreterLib::Position position;
	GSInterpreterLib *interpreter;

	GLEFileInfo thisFile;


	if (caller == CalledByObject)
		thisFile = objFile;
	else
		thisFile = gleFile;

	// This is a fudge:
	qDebug() << "Waiting for end of rendering\n";
	// I've commented out the following loop for now ...
	// It appears that it also works fine without it
	// Should this be replaced by QWaitCondition later on?
	// while(isRendering);
	qDebug() << "Rendering available\n";

	mutex.lock();
	isRendering = true;
	mutex.unlock();

	QFile file(thisFile.epsFile());
	// Check that the file can be opened and scan through for
	// BBOX and file size
	if (file.open( QIODevice::ReadOnly ))
	{
		// Get the file size
		position.first = 0;
		position.second = file.size();

		// Scan through file for BoundingBox and set width and height
		QTextStream in(&file);
		while (!in.atEnd())
		{
			QString line = in.readLine();
			if (rxbb.exactMatch(line))
			{
				b1 = rxbb.capturedTexts()[1].toInt();
				b2 = rxbb.capturedTexts()[2].toInt();
				b3 = rxbb.capturedTexts()[3].toInt();
				b4 = rxbb.capturedTexts()[4].toInt();
				break;
			}
		}

		// Derive the width and height and close the file
		width = b3-b1+1;
		height = b4-b2+1;
		file.close();

		// Auto scale new files to size of drawing area
		if (shouldAutoScale()) {
			setAutoScale(false);
			setDPI(QGLE::computeAutoScaleDPIFromPts(drawingAreaSize, width, height));
		}

		// Open the file for reading by GS
		FILE *fh = fopen(thisFile.epsFile().toLatin1().constData(), "r");

		// Create the interpreter object and set a few parameters
		interpreter = new GSInterpreterLib();
		interpreter->setOrigin(b1, b2);
		interpreter->setProgressive(false);
		interpreter->setMedia("a4");
		interpreter->setMagnify(1);
		interpreter->setOrientation(0);
		interpreter->setDPI(dpi);
		int img_wd = (int)floor((double)dpi/72.0*width+1);
		int img_hi = (int)floor((double)dpi/72.0*height+1);
		interpreter->setSize(img_wd, img_hi);
		interpreter->setAABits(4,4);

		qRegisterMetaType<GLEFileInfo>("GLEFileInfo");
		qRegisterMetaType<QImage>("QImage");
		connect(interpreter, SIGNAL(Finished(QImage)),
				this, SLOT(prepareFileInfo(QImage)));
		storedCaller = caller;
		storedObject = obj;

		// Run the interpreter
		qDebug() << "Starting the renderer\n";
		interpreter->run(fh, position);

		if (interpreter->hasError()) {
			const GSError& e = interpreter->getError();
			QString error = QString("Ghostscript error: %1 %2")
			                        .arg(e.getCode())
			                        .arg(e.getName());
			qDebug() << "Error: " << error;
			emit serverError(error);
		}

		GhostScriptOutput = interpreter->getOutpuStreamText();
		GhostScriptError = interpreter->getErrorStreamText();

		// Clean up
		fclose(fh);
		delete interpreter;
	}
}

void GLEServerThread::EPSToImage(QString file, double new_dpi, bool autoScale)
{
	gleFile.clear();
	gleFile.setEpsFile(file);
	dpi = new_dpi;
	setAutoScale(autoScale);
	qDebug() << "EPS Rendering " << file;
	render(CalledByQGLE);
}

void GLEServerThread::PSToImage(QString postscriptCode, double new_dpi, GLEDrawingObject *obj)
{
	dpi = new_dpi;
	GLEDrawObject* gleobj = (GLEDrawObject*)obj->getGLEObject();

	// Get the bounding box (important!)
	GLERectangle rect;
	gleobj->getPSBoundingBox(&rect);
	double width  = 72*rect.getXMax()/CM_PER_INCH+2;
	double height = 72*rect.getYMax()/CM_PER_INCH+2;

	// Create the interpreter object and set a few parameters
	GSInterpreterLib *interpreter = new GSInterpreterLib();
	interpreter->setOrigin(0, 0);
	interpreter->setProgressive(false);
	interpreter->setMedia("a4");
	interpreter->setMagnify(1);
	interpreter->setOrientation(0);
	interpreter->setDPI(new_dpi);
	interpreter->setAlpha(true);
	int img_wd = (int)floor((double)new_dpi/72.0*width+1);
	int img_hi = (int)floor((double)new_dpi/72.0*height+1);
	interpreter->setSize(img_wd, img_hi);
	interpreter->setAABits(4,4);

	qRegisterMetaType<GLEFileInfo>("GLEFileInfo");
	qRegisterMetaType<QImage>("QImage");
	connect(interpreter, SIGNAL(Finished(QImage)),
	        this, SLOT(prepareFileInfo(QImage)));
	storedCaller = CalledByObject;
	storedObject = obj;

	// Run the interpreter
	qDebug() << "Starting the renderer\n";
	interpreter->run(postscriptCode);

	if (interpreter->hasError()) {
		const GSError& e = interpreter->getError();
		QString error = QString("Ghostscript error: %1 %2")
		                        .arg(e.getCode())
		                        .arg(e.getName());
		qDebug() << "Error: " << error;
		emit serverError(error);
	}

	GhostScriptOutput = interpreter->getOutpuStreamText();
	GhostScriptError = interpreter->getErrorStreamText();

	// Clean up
	delete interpreter;
}

void GLEServerThread::prepareFileInfo(QImage image)
{
	// QMutex::lock: Deadlock detected in thread -1217720192
	// mutex.lock();
	if (storedCaller == CalledByQGLE)
	{
		emit readyToGo(image);
	}
	else if (storedCaller == CalledByObject)
	{
		connect(this, SIGNAL(objectRenderingComplete(QImage)),
				storedObject, SLOT(imageRendered(QImage)));
		emit objectRenderingComplete(image);
		disconnect(this, SIGNAL(objectRenderingComplete(QImage)),
				storedObject, SLOT(imageRendered(QImage)));
	}
	else if (storedCaller == CalledBySocket)
	{
		gleFile.setTemp(false);
		gleFile.setPreviewModeImage(image,dpi);
		gleFile.setEditModeImage(image);
		gleFile.updateTimeStamp();
		emit imageReady(gleFile);
	}
	else
	{
		qDebug() << "SHOULD NEVER GET HERE!";
	}
	qDebug() << "Rendering complete\n";
	isRendering = false;
	// mutex.unlock();


}
