/***********************************************************************************
 * QGLE - A Graphical Interface to GLE                                             *
 * Copyright (C) 2006  A. S. Budden & J. Struyf                                    *
 *                                                                                 *
 * This program is free software; you can redistribute it and/or                   *
 * modify it under the terms of the GNU General Public License                     *
 * as published by the Free Software Foundation; either version 2                  *
 * of the License, or (at your option) any later version.                          *
 *                                                                                 *
 * This program is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   *
 * GNU General Public License for more details.                                    *
 *                                                                                 *
 * You should have received a copy of the GNU General Public License               *
 * along with this program; if not, write to the Free Software                     *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. *
 *                                                                                 *
 * Also add information on how to contact you by electronic and paper mail.        *
 ***********************************************************************************/

#include <QtDebug>
#include "text.h"
#include "qgle_statics.h"
#include "gledrawing.h"

// The constructor for the line object
GLEText::GLEText(double resolution, QSize imageSize, QObject *parent) :
	GLEDrawingObject(resolution, imageSize, parent)
{
	lastdpi = -1;
	psCode = "";
	// Make sure the line is updated if a point changes or
	// the image size changes
	connect(this, SIGNAL(pointChanged()),
			this, SLOT(updateText()));
	connect(this, SIGNAL(imageChanged()),
			this, SLOT(resChanged()));
	connect(this, SIGNAL(propertyChanged(int)),
			this, SLOT(updateFromProperty(int)));

	amove = false;
	haveImage = false;
	drawPoint = QPointF(0.0,0.0);

	// More to be added
	validProperties
		<< Text
		<< FontName
		<< FontSize
		<< Alignment
		<< LineColour;
	properties[Text] = propertyDescriptions[Text].defaultValue;
	properties[FontName] = propertyDescriptions[FontName].defaultValue;
	properties[FontSize] = propertyDescriptions[FontSize].defaultValue;
	properties[Alignment] = propertyDescriptions[Alignment].defaultValue;
	properties[LineColour] = propertyDescriptions[LineColour].defaultValue;
}

void GLEText::resChanged()
{
	if (dpi != lastdpi)
	{
		qDebug() << "Resolution: " << dpi;
		if (haveImage)
			emit renderPostscript(psCode, dpi, this);
		dpi = lastdpi;
	}
	else
	{
		updateText();
	}
}

void GLEText::createOTracks()
{
}

void GLEText::updateFromPropertyNoDirty(int property)
{
	GLETextDO* text = (GLETextDO*)getGLEObject();
	GLEPropertyStore* text_settings = text->getProperties();

	switch(property)
	{
		case Text:
		case LineColour:
		case FontName:
		case FontSize: {
			// Font is
			// fontList[properties[FontName].value<QString>]
			// (default "rm")
			int font = getProperty(FontName).toInt();
			QColor color = getProperty(LineColour).value<QColor>();
			GLEColor* gle_color = text_settings->getColorProperty(GLEDOPropertyColor);
			gle_color->setRGB255(color.red(), color.green(), color.blue());
			text_settings->setFontProperty(GLEDOPropertyFont, gleInterface->getFont(font));
			text_settings->setIntProperty(GLEDOPropertyFontStyle, GLEFontStyleRoman);
			text_settings->setRealProperty(GLEDOPropertyFontSize, getProperty(FontSize).toDouble());
			text->setText(getProperty(Text).toString().toUtf8());
			text->render(gleInterface);

			// Get the bounding box (important!)
			text->getPSBoundingBox(&rect);

			psCode = QString("");
			psCode += gleInterface->getInitialPostScript();
			psCode += text->getPostScriptCode();
			psCode += gleInterface->getTerminatePostScript();

			// Now we need to pass this to the ghostscript renderer...
			emit renderPostscript(psCode, dpi, this);

			break;
		}
		case Alignment: {
			text_settings->setIntProperty(GLEDOPropertyJustify, getProperty(GLEText::Alignment).toInt());
			updateText();
			break;
		}
		default:
			updateText();
	}

	// updateText() is called after rendering is complete.
}

void GLEText::updateFromProperty(int property)
{
	updateFromPropertyNoDirty(property);
	((GLEDrawingArea*)parent())->setDirtyAndSave();
}

// Update the painter path
void GLEText::updateText()
{
	// Only do this if both the start and end have been set
	GLETextDO* text = (GLETextDO*)getGLEObject();
	if (isSet(ReferencePoint) && text != NULL)
	{
		// This is guaranteed to have been created in the constructor
		// of GLEDrawingObject
		delete(paintPath);
		// Create a new path starting at lineStart
		paintPath = new QPainterPath();
		QPointF qtText = getQtPoint(ReferencePoint);
		QPointF qtRelative;
		QPointF refPt;

		// Update GLE object
		text->setPosition(QGLE::QPointFToGLEPoint(getGLEPoint(ReferencePoint)));
		text->setText(getProperty(Text).toString().toUtf8());

		int alignProp = getProperty(Alignment).toInt();

		// Now work out the point at which the object is drawn:
		switch(alignProp)
		{
			case GLEJustifyTL:
				refPt = QPointF(0.0, -rect.getYMax());
				break;

			case GLEJustifyTR:
				refPt = QPointF(rect.getXMax(), -rect.getYMax());
				break;

			case GLEJustifyBL:
				refPt = QPointF(0.0, 0.0);
				break;

			case GLEJustifyBR:
				refPt = QPointF(rect.getXMax(), 0.0);
				break;

			case GLEJustifyLC:
				refPt = QPointF(0.0, -rect.getYMax()/2.0);
				break;

			case GLEJustifyRC:
				refPt = QPointF(rect.getXMax(), -rect.getYMax()/2.0);
				break;

			case GLEJustifyTC:
				refPt = QPointF(rect.getXMax()/2.0, -rect.getYMax());
				break;

			case GLEJustifyBC:
				refPt = QPointF(rect.getXMax()/2.0, 0.0);
				break;

			case GLEJustifyLeft:
				refPt = QPointF(0.0, -text->getBaseLine());
				break;

			case GLEJustifyRight:
				refPt = QPointF(rect.getXMax(), -text->getBaseLine());
				break;

			case GLEJustifyCenter:
				refPt = QPointF(rect.getXMax()/2.0, -text->getBaseLine());
				break;

			case GLEJustifyCC:
			default: // Case 0: CC
				// Centre point is at rect.getXMax()/2, rect.getYMax()/2
				// in GLE coordinates.  Therefore, the point at which to
				// draw the pixmap is the reference point minus the
				// X & Y values in Qt coordinates...
				refPt = QPointF(rect.getXMax()/2.0,
						-rect.getYMax()/2.0);
		}
		qtRelative = QGLE::relGLEToQt(refPt, dpi);
		QPointF offset(-GS_OFFSET*dpi,GS_OFFSET*dpi-renderedText.height());
		drawPoint = qtText - qtRelative + offset;

		// This needs to be changed to use the bounding box
		// of the string
		QSizeF qtSize;
		qtSize.setWidth(QGLE::relGLEToQt(rect.getXMax(), dpi));
		qtSize.setHeight(QGLE::relGLEToQt(rect.getYMax(), dpi));
		// The space from the top of the image is the following:
		double yoffs = renderedText.height()-rect.getYMax()*dpi/CM_PER_INCH-GS_OFFSET*dpi;
		paintPath->addRect(QRectF(drawPoint.x()+GS_OFFSET*dpi,
					drawPoint.y()+yoffs,
					qtSize.width(),
					qtSize.height()));

		osnapHandles.clear();
		osnapHandles.append(QPair<QPointF,int>(qtText,QGLE::CentreSnap));
	}
}

void GLEText::imageRendered(QImage newImage)
{
	renderedText = QPixmap::fromImage(newImage);
	//newImage.save("/home/al/test.png","png",-1);
	QGLE::flushIO();
	haveImage = true;
	updateText();
}

void GLEText::addRelativeOSnaps(QPointF p)
{
	UNUSED_ARG(p);
}

void GLEText::draw(QPainter *p)
{
	// This needs to be changed to plot an image object
	if (haveImage)
		p->drawPixmap(drawPoint, renderedText);
	// QGLE::drawCross(p, getQtPoint(ReferencePoint), AMOVE_LENGTH, pen());
}

QList<QPointF> GLEText::getTangents(QPointF p)
{
	QList<QPointF> empty;
	UNUSED_ARG(p);
	return(empty);
}

bool GLEText::hasTangents()
{
	return(false);
}

QList<QPointF> GLEText::getPerpendiculars(QPointF p)
{
	QList<QPointF> empty;
	UNUSED_ARG(p);
	return(empty);
}

bool GLEText::hasPerpendiculars()
{
	return(false);
}




double GLEText::distanceToPoint(QPointF p, QPointF *nearestPoint)
{
	// Probably change this to work off the bounding box of the text
	if (paintPath->contains(p))
	{
		if (nearestPoint)
			*nearestPoint = p;
		return(0.0);
	}
	else
	{
		// TODO: This is wrong! Find the distance to the edge of the
		// bounding box!
		QPointF amovePoint = getQtPoint(ReferencePoint);
		if (nearestPoint)
			*nearestPoint = amovePoint;
		return(QGLE::distance(amovePoint, p));
	}
}

// Set a point (start or end in the case of a line)
void GLEText::setPoint(int pointChoice, QPointF p)
{
	switch(pointChoice)
	{
		case ReferencePoint:
			pointHash[pointChoice] = QGLE::absQtToGLE(p, dpi, pixmap.height());
			break;
	}
	updateText();
}

// Set a point, but don't update object
void GLEText::setPointNoUpdate(int pointChoice, QPointF p)
{
	switch(pointChoice)
	{
		case ReferencePoint:
			pointHash[pointChoice] = QGLE::absQtToGLE(p, dpi, pixmap.height());
			break;
	}
}

QList<QPointF> GLEText::intersections(double qtm, double qtc, bool vertical)
{
	UNUSED_ARG(qtm);
	UNUSED_ARG(qtc);
	UNUSED_ARG(vertical);
	return QList<QPointF>();
}

QList<QPointF> GLEText::intersections(QPointF qtp1, QPointF qtp2)
{
	UNUSED_ARG(qtp1);
	UNUSED_ARG(qtp2);
	return QList<QPointF>();
}

QList<QPointF> GLEText::intersections(QPointF qtp1, double angle)
{
	UNUSED_ARG(qtp1);
	UNUSED_ARG(angle);
	return QList<QPointF>();

}

bool GLEText::isInside(QPointF p)
{
	return(paintPath->contains(p));
}


void GLEText::scaleBy(double xratio, double yratio)
{
	// TODO: Also scale the text size?
	pointHash.clear();

	QPointF oldOffset = getQtPoint(ReferencePoint, true) - basePoint;
	QPointF newOffset;
	newOffset.setX(oldOffset.x() * xratio);
	newOffset.setY(oldOffset.y() * yratio);

	setPoint(ReferencePoint, basePoint + newOffset);
}

void GLEText::moveBy(QPointF offset)
{
	pointHash.clear();
	// Moves relative to storedPointHash
	QPointF start = getQtPoint(ReferencePoint, true);
	setPoint(ReferencePoint, start + offset);
}

void GLEText::rotateBy(double radians)
{
	// TODO: Also rotate the text.
	pointHash.clear();
	// Moves relative to storedPointHash
	QPointF start = getQtPoint(ReferencePoint, true);
	setPoint(ReferencePoint, QGLE::rotateAboutPoint(start, radians, basePoint));
}
